/*-
 * Copyright (c) 2003 Hiroyuki Aizu <aizu@navi.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
#include "opt_acpi.h"
#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/bus.h>
#include <sys/power.h>

#include "acpi.h"
#include <dev/acpica/acpivar.h>

#define _COMPONENT      ACPI_TOSHIBA
ACPI_MODULE_NAME("TOSHIBA")

/* Toshiba HCI interface definitions
 *
 * HCI is Toshiba's "Hardware Control Interface" which is supposed to
 * be uniform across all their models.  Ideally we would just call
 * dedicated ACPI methods instead of using this primitive interface.
 * However the ACPI methods seem to be incomplete in some areas (for
 * example they allow setting, but not reading, the LCD brightness value),
 * so this is still useful.
 */

#define METHOD_HCI		"GHCI"
#define METHOD_HCI_ENABLE	"ENAB"

#define HCI_WORDS			6

/* operations */
#define HCI_SET				0xFF00
#define HCI_GET				0xFE00

/* return codes */
#define HCI_SUCCESS			0x0000
#define HCI_FAILURE			0x1000
#define HCI_NOT_SUPPORTED		0x8000
#define HCI_EMPTY			0x8C00

/* registers */
#define HCI_REG_LCD_BACKLIGHT		0x0002
#define HCI_REG_FAN			0x0004
#define HCI_REG_SYSTEM_EVENT		0x0016
#define HCI_REG_VIDEO_OUTPUT		0x001C
#define HCI_REG_HOTKEY_EVENT		0x001E
#define HCI_REG_LCD_BRIGHTNESS		0x002A

/* field definitions */
#define HCI_LCD_BRIGHTNESS_BITS		3
#define HCI_LCD_BRIGHTNESS_SHIFT	(16-HCI_LCD_BRIGHTNESS_BITS)
#define HCI_LCD_BRIGHTNESS_LEVELS	(1 << HCI_LCD_BRIGHTNESS_BITS)
#define HCI_VIDEO_OUTPUT_LCD		0x1
#define HCI_VIDEO_OUTPUT_CRT		0x2
#define HCI_VIDEO_OUTPUT_TV		0x4

struct hci_regs {
    UINT32 reg[HCI_WORDS];
#define HCI_REG_AX 0
#define HCI_REG_BX 1
#define HCI_REG_CX 2
#define HCI_REG_DX 3
#define HCI_REG_SI 4
#define HCI_REG_DI 5
};

struct acpi_toshiba_softc {
    device_t		dev;
    ACPI_HANDLE		handle;

    struct sysctl_ctx_list	toshiba_sysctl_ctx;
    struct sysctl_oid		*toshiba_sysctl_tree;
};

static int acpi_toshiba_probe(device_t);
static int acpi_toshiba_attach(device_t);

static int acpi_toshiba_sysctl_force_fan(SYSCTL_HANDLER_ARGS);
static int acpi_toshiba_sysctl_video_output(SYSCTL_HANDLER_ARGS);
static int acpi_toshiba_sysctl_lcd_brightness(SYSCTL_HANDLER_ARGS);
static int acpi_toshiba_sysctl_lcd_backlight(SYSCTL_HANDLER_ARGS);

static ACPI_STATUS acpi_toshiba_fn_extension_enable(device_t, ACPI_HANDLE);
static ACPI_STATUS get_hci_fn_key(ACPI_HANDLE, UINT32 *);
static ACPI_STATUS set_hci_force_fan(ACPI_HANDLE, UINT32);
static ACPI_STATUS get_hci_force_fan(ACPI_HANDLE, UINT32 *);
static ACPI_STATUS set_hci_video_output(ACPI_HANDLE, UINT32);
static ACPI_STATUS get_hci_video_output(ACPI_HANDLE, UINT32 *);
static ACPI_STATUS set_hci_lcd_brightness(ACPI_HANDLE, UINT32);
static ACPI_STATUS get_hci_lcd_brightness(ACPI_HANDLE, UINT32 *);
static ACPI_STATUS set_hci_lcd_backlight(ACPI_HANDLE, UINT32);
static ACPI_STATUS get_hci_lcd_backlight(ACPI_HANDLE, UINT32 *);

static ACPI_STATUS call_hci(ACPI_HANDLE, struct hci_regs *, struct hci_regs *);

static void fan_next(ACPI_HANDLE);
static void video_output_next(ACPI_HANDLE);
static void lcd_brightness_up(ACPI_HANDLE);
static void lcd_brightness_down(ACPI_HANDLE);
static void lcd_backlight_next(ACPI_HANDLE);

static void system_event(device_t, ACPI_HANDLE);

static void acpi_toshiba_notify_handler(ACPI_HANDLE, UINT32, void *);
static void acpi_toshiba_button_notify_pressed_for_sleep(void *);
static void acpi_toshiba_button_pressed_for_power_profile(void *);

static UINT32 hci_video_output_mask = 0;

#define FN_EXT_DISABLE	0
#define FN_EXT_ENABLE	1
static int enable_fn_ext = FN_EXT_ENABLE;
TUNABLE_INT("hw.acpi.toshiba.enable_fn_extension", &enable_fn_ext);


/*
 * Device VALD
 * HID      Model
 * -------------------------------------
 * TOS6200  Libretto L Series
 * TOS6207  Dynabook SS2110 Series
 */
static int
acpi_toshiba_probe(device_t dev)
{
    if (!acpi_disabled("toshiba")) {
        if (acpi_get_type(dev) == ACPI_TYPE_DEVICE) {
            if (acpi_MatchHid(dev, "TOS6200") ||
                acpi_MatchHid(dev, "TOS6207")) {
                device_set_desc(dev, "Toshiba Notebook Extra HCI driver");
                return (0);
            }
        }
    }
    return(ENXIO);
}

static int
acpi_toshiba_attach(device_t dev)
{
    UINT32 val;
    struct acpi_softc		*acpi_sc;
    struct acpi_toshiba_softc	*sc;

    if ((sc = device_get_softc(dev)) == NULL)
        return(ENXIO);

    sc->dev = dev;
    sc->handle = acpi_get_handle(dev);
    if (sc->handle == NULL) {
        return(ENXIO);
    }

    acpi_sc = acpi_device_get_parent_softc(dev);
    sysctl_ctx_init(&sc->toshiba_sysctl_ctx);

    sc->toshiba_sysctl_tree = SYSCTL_ADD_NODE(&sc->toshiba_sysctl_ctx,
        SYSCTL_CHILDREN(acpi_sc->acpi_sysctl_tree), OID_AUTO,
        "toshiba", CTLFLAG_RD, 0, "");

    /* Setup sysctl tree */
    if (ACPI_SUCCESS(get_hci_force_fan(sc->handle, &val))) {
        SYSCTL_ADD_PROC(&sc->toshiba_sysctl_ctx,
            SYSCTL_CHILDREN(sc->toshiba_sysctl_tree), OID_AUTO,
            "force_fan", CTLTYPE_INT | CTLFLAG_RW | CTLFLAG_ANYBODY,
            sc, 0, acpi_toshiba_sysctl_force_fan, "I", "");
    }

    if (ACPI_SUCCESS(get_hci_video_output(sc->handle, &val))) {
        if (ACPI_FAILURE(set_hci_video_output(sc->handle, val))) {
            hci_video_output_mask = 0x0100;
        }

        SYSCTL_ADD_PROC(&sc->toshiba_sysctl_ctx,
            SYSCTL_CHILDREN(sc->toshiba_sysctl_tree), OID_AUTO,
            "video_output", CTLTYPE_INT | CTLFLAG_RW | CTLFLAG_ANYBODY,
            sc, 0, acpi_toshiba_sysctl_video_output, "I", "");
    }

    if (ACPI_SUCCESS(get_hci_lcd_brightness(sc->handle, &val))) {
        SYSCTL_ADD_PROC(&sc->toshiba_sysctl_ctx,
            SYSCTL_CHILDREN(sc->toshiba_sysctl_tree), OID_AUTO,
            "lcd_brightness_level", CTLTYPE_INT | CTLFLAG_RW | CTLFLAG_ANYBODY,
            sc, 0, acpi_toshiba_sysctl_lcd_brightness, "I", "");
    }

    if (ACPI_SUCCESS(get_hci_lcd_backlight(sc->handle, &val))) {
        SYSCTL_ADD_PROC(&sc->toshiba_sysctl_ctx,
            SYSCTL_CHILDREN(sc->toshiba_sysctl_tree), OID_AUTO,
            "lcd_backlight", CTLTYPE_INT | CTLFLAG_RW | CTLFLAG_ANYBODY,
            sc, 0, acpi_toshiba_sysctl_lcd_backlight, "I", "");
    }

    /* Enable Fn-key extension */
    if (enable_fn_ext == FN_EXT_ENABLE) {
        acpi_toshiba_fn_extension_enable(dev, sc->handle);
    }

    return (0);
}


static int
acpi_toshiba_sysctl_force_fan(SYSCTL_HANDLER_ARGS)
{
    struct acpi_toshiba_softc *sc;
    UINT32 arg;
    int error = 0;

    sc = (struct acpi_toshiba_softc *)oidp->oid_arg1;
    if (ACPI_FAILURE(get_hci_force_fan(sc->handle, &arg))) {
        return (EINVAL);
    }

    error = sysctl_handle_int(oidp, &arg, 0, req);

    /* error or no new value */
    if ((error != 0) || (req->newptr == NULL)) {
        return (error);
    }

    if ((arg < 0) || (arg > 1)) {
        return (EINVAL);
    }

    if (ACPI_FAILURE(set_hci_force_fan(sc->handle, arg))) {
        return (EINVAL);
    }

    return (0);
}


static int
acpi_toshiba_sysctl_video_output(SYSCTL_HANDLER_ARGS)
{
    struct acpi_toshiba_softc *sc;
    UINT32 arg;
    int error = 0;

    sc = (struct acpi_toshiba_softc *)oidp->oid_arg1;
    if (ACPI_FAILURE(get_hci_video_output(sc->handle, &arg))) {
        return(EINVAL);
    }

    error = sysctl_handle_int(oidp, &arg, 0, req);

    /* error or no new value */
    if ((error != 0) || (req->newptr == NULL)) {
        return (error);
    }

    /* range check */
    if ((arg < 0) || (arg > 7)) {
        return (EINVAL);
    }

    if (ACPI_FAILURE(set_hci_video_output(sc->handle, arg))) {
        return (EINVAL);
    }

    return (0);
}

static int
acpi_toshiba_sysctl_lcd_brightness(SYSCTL_HANDLER_ARGS)
{
    struct acpi_toshiba_softc *sc;
    UINT32 arg;
    int error = 0;

    sc = (struct acpi_toshiba_softc *)oidp->oid_arg1;

    if (ACPI_FAILURE(get_hci_lcd_brightness(sc->handle, &arg))) {
        return (EINVAL);
    }

    error = sysctl_handle_int(oidp, &arg, 0, req);

    /* error or no new value */
    if ((error != 0) || (req->newptr == NULL)) {
        return (error);
    }

    /* range check */
    if ((arg < 0) || (arg > 7)) {
        return (EINVAL);
    }

    if (ACPI_FAILURE(set_hci_lcd_brightness(sc->handle, arg))) {
        return (EINVAL);
    }

    return (0);
}

static int
acpi_toshiba_sysctl_lcd_backlight(SYSCTL_HANDLER_ARGS)
{
    struct acpi_toshiba_softc *sc;
    UINT32 arg;
    int error = 0;

    sc = (struct acpi_toshiba_softc *)oidp->oid_arg1;

    if (ACPI_FAILURE(get_hci_lcd_backlight(sc->handle, &arg))) {
        return (EINVAL);
    }

    error = sysctl_handle_int(oidp, &arg, 0, req);

    /* error or no new value */
    if ((error != 0) || (req->newptr == NULL)) {
        return (error);
    }

    /* range check */
    if ((arg < 0) || (arg > 1)) {
        return (EINVAL);
    }

    if (ACPI_FAILURE(set_hci_lcd_backlight(sc->handle, arg))) {
        return (EINVAL);
    }

    return (0);
}

static ACPI_STATUS
acpi_toshiba_fn_extension_enable(device_t dev, ACPI_HANDLE handle)
{
    ACPI_STATUS status;

    status = AcpiEvaluateObject(handle, METHOD_HCI_ENABLE, NULL, NULL);
    if (ACPI_FAILURE(status)) {
        device_printf(dev, "acpi_toshiba_ex_enable() evaluate failed\n");
        return (status);
    }

    /* Install Notify handler */
    AcpiInstallNotifyHandler(handle, ACPI_DEVICE_NOTIFY,
                             acpi_toshiba_notify_handler, dev);

    return (status);
}


ACPI_STATUS
get_hci_fn_key(ACPI_HANDLE handle, UINT32 *key)
{
    struct hci_regs args, results;

    bzero(&args, sizeof(struct hci_regs));

    args.reg[HCI_REG_AX] = HCI_GET;
    args.reg[HCI_REG_BX] = HCI_REG_SYSTEM_EVENT;

    if (ACPI_FAILURE(call_hci(handle, &args, &results))) {
        return (0);
    }

    if (results.reg[HCI_REG_AX] != HCI_SUCCESS) {
        return (AE_ERROR);
    }

    *key = results.reg[HCI_REG_CX];
    return (AE_OK);
}

ACPI_STATUS
set_hci_force_fan(ACPI_HANDLE handle, UINT32 state)
{
    struct hci_regs args, results;

    bzero(&args, sizeof(struct hci_regs));
    args.reg[HCI_REG_AX] = HCI_SET;
    args.reg[HCI_REG_BX] = HCI_REG_FAN;
    args.reg[HCI_REG_CX] = (state << 7);

    if (ACPI_FAILURE(call_hci(handle, &args, &results))) {
        return (AE_ERROR);
    }

    if (results.reg[HCI_REG_AX] != HCI_SUCCESS) {
        return (AE_ERROR);
    }

    return (AE_OK);
}


ACPI_STATUS
get_hci_force_fan(ACPI_HANDLE handle, UINT32 *fan)
{
    struct hci_regs args, results;

    bzero(&args, sizeof(struct hci_regs));
    args.reg[HCI_REG_AX] = HCI_GET;
    args.reg[HCI_REG_BX] = HCI_REG_FAN;

    if (ACPI_FAILURE(call_hci(handle, &args, &results))) {
        return (AE_ERROR);
    }

    if (results.reg[HCI_REG_AX] != HCI_SUCCESS) {
        *fan = 0;
        return (AE_ERROR);
    }

    *fan = (results.reg[HCI_REG_CX] >> 7);
    return (AE_OK);
}

ACPI_STATUS
set_hci_video_output(ACPI_HANDLE handle, UINT32 video_output)
{
    struct hci_regs args, results;

    bzero(&args, sizeof(struct hci_regs));
    args.reg[HCI_REG_AX] = HCI_SET;
    args.reg[HCI_REG_BX] = HCI_REG_VIDEO_OUTPUT;
    args.reg[HCI_REG_CX] = (video_output | hci_video_output_mask);

    if (ACPI_FAILURE(call_hci(handle, &args, &results))) {
        return (AE_ERROR);
    }

    if (results.reg[HCI_REG_AX] != HCI_SUCCESS) {
        return (AE_ERROR);
    }

    return (AE_OK);
}

ACPI_STATUS
get_hci_video_output(ACPI_HANDLE handle, UINT32 *video_out)
{
    struct hci_regs args, results;

    bzero(&args, sizeof(struct hci_regs));
    args.reg[HCI_REG_AX] = HCI_GET;
    args.reg[HCI_REG_BX] = HCI_REG_VIDEO_OUTPUT;

    if (ACPI_FAILURE(call_hci(handle, &args, &results))) {
        return (AE_ERROR);
    }

    if (results.reg[HCI_REG_AX] != HCI_SUCCESS) {
        *video_out = 0;
        return (AE_ERROR);
    }

    *video_out = (results.reg[HCI_REG_CX] & 0x00FF);

    return (AE_OK);
}

ACPI_STATUS
set_hci_lcd_brightness(ACPI_HANDLE handle, UINT32 brightness)
{
    struct hci_regs args, results;

    bzero(&args, sizeof(struct hci_regs));
    args.reg[HCI_REG_AX] = HCI_SET;
    args.reg[HCI_REG_BX] = HCI_REG_LCD_BRIGHTNESS;
    args.reg[HCI_REG_CX] = (brightness << HCI_LCD_BRIGHTNESS_SHIFT);

    if (ACPI_FAILURE(call_hci(handle, &args, &results))) {
        return (AE_ERROR);
    }

    if (results.reg[HCI_REG_AX] != HCI_SUCCESS) {
        return (AE_ERROR);
    }

    return (AE_OK);
}

ACPI_STATUS
get_hci_lcd_brightness(ACPI_HANDLE handle, UINT32 *brightness)
{
    struct hci_regs args, results;

    bzero(&args, sizeof(struct hci_regs));
    args.reg[HCI_REG_AX] = HCI_GET;
    args.reg[HCI_REG_BX] = HCI_REG_LCD_BRIGHTNESS;

    if (ACPI_FAILURE(call_hci(handle, &args, &results))) {
        return (AE_ERROR);
    }

    if (results.reg[HCI_REG_AX] != HCI_SUCCESS) {
        *brightness = 0;
        return (AE_ERROR);
    }

    *brightness = (results.reg[HCI_REG_CX] >> HCI_LCD_BRIGHTNESS_SHIFT) & 0x07;
    return (AE_OK);
}

ACPI_STATUS
set_hci_lcd_backlight(ACPI_HANDLE handle, UINT32 brightness)
{
    struct hci_regs args, results;

    bzero(&args, sizeof(struct hci_regs));
    args.reg[HCI_REG_AX] = HCI_SET;
    args.reg[HCI_REG_BX] = HCI_REG_LCD_BACKLIGHT;
    args.reg[HCI_REG_CX] = brightness;

    if (ACPI_FAILURE(call_hci(handle, &args, &results))) {
        return (AE_ERROR);
    }

    if (results.reg[HCI_REG_AX] != HCI_SUCCESS) {
        return (AE_ERROR);
    }

    return (AE_OK);
}

ACPI_STATUS
get_hci_lcd_backlight(ACPI_HANDLE handle, UINT32 *brightness)
{
    struct hci_regs args, results;

    bzero(&args, sizeof(struct hci_regs));
    args.reg[HCI_REG_AX] = HCI_GET;
    args.reg[HCI_REG_BX] = HCI_REG_LCD_BACKLIGHT;

    if (ACPI_FAILURE(call_hci(handle, &args, &results))) {
        return (AE_ERROR);
    }

    if (results.reg[HCI_REG_AX] != HCI_SUCCESS) {
        *brightness = 0;
        return (AE_ERROR);
    }

    *brightness = results.reg[HCI_REG_CX];
    return (AE_OK);
}


ACPI_STATUS
call_hci(ACPI_HANDLE h, struct hci_regs *in, struct hci_regs *out)
{
    ACPI_OBJECT_LIST args;
    ACPI_BUFFER results;
    ACPI_OBJECT obj[HCI_WORDS];
    ACPI_OBJECT *res, *tmp;
    ACPI_STATUS status = AE_OK;
    int i;

    args.Count = HCI_WORDS;
    args.Pointer = obj;
    for (i = 0; i < HCI_WORDS; ++i) {
        obj[i].Type = ACPI_TYPE_INTEGER;
        obj[i].Integer.Value = in->reg[i];
    }

    results.Pointer = NULL;
    results.Length = ACPI_ALLOCATE_BUFFER;

    if (ACPI_FAILURE(AcpiEvaluateObject(h, METHOD_HCI, &args, &results))) {
        goto end;
    }

    res = (ACPI_OBJECT *)results.Pointer;
    if (res == NULL) {
        status = AE_ERROR;
        goto end;
    }

    if ((res->Type != ACPI_TYPE_PACKAGE) || (res->Package.Count > HCI_WORDS)) {
        status = AE_ERROR;
        goto end;
    }
    
    for (i=0; i<HCI_WORDS; i++) {
        tmp = &res->Package.Elements[i];
        out->reg[i] = tmp->Integer.Value;
    }

end:
    if (results.Pointer != NULL)
        AcpiOsFree(results.Pointer);

    return(status);
}


void
fan_next(ACPI_HANDLE handle)
{
    UINT32 fan_status;

    if (ACPI_FAILURE(get_hci_force_fan(handle, &fan_status))) {
        return;
    }
    fan_status++;
    if (fan_status > 1) {
        fan_status = 0;
    }

    set_hci_force_fan(handle, fan_status);
}

void
video_output_next(ACPI_HANDLE handle)
{
    UINT32 video_out;

    if (ACPI_FAILURE(get_hci_video_output(handle, &video_out))) {
        return;
    }

    video_out++;
    
    set_hci_video_output(handle, video_out);
}

void
lcd_brightness_up(ACPI_HANDLE handle)
{
    UINT32 brightness;

    if (ACPI_FAILURE(get_hci_lcd_brightness(handle, &brightness))) {
        return;
    }

    brightness++;
    if (brightness > 7) {
        return;
    }
    
    set_hci_lcd_brightness(handle, brightness);
}

void
lcd_brightness_down(ACPI_HANDLE handle)
{
    UINT32 brightness;

    if (ACPI_FAILURE(get_hci_lcd_brightness(handle, &brightness))) {
        return;
    }
    if (brightness == 0) {
        return;
    }

    brightness--;
    set_hci_lcd_brightness(handle, brightness);
}

void
lcd_backlight_next(ACPI_HANDLE handle)
{
    UINT32 status;

    if (ACPI_FAILURE(get_hci_lcd_backlight(handle, &status))) {
        return;
    }

    status++;
    if (status > 1) {
        status = 0;
    }

    set_hci_lcd_backlight(handle, status);
}

#define FN_F_PRESS	0x0121
#define FN_F_RELEASE	0x01A1

#define FN_F1_PRESS	0x013B
#define FN_F1_RELEASE	0x01BB
#define FN_F2_PRESS	0x013C
#define FN_F2_RELEASE	0x01BC
#define FN_F3_PRESS	0x013D
#define FN_F3_RELEASE	0x01BD
#define FN_F4_PRESS	0x013E
#define FN_F4_RELEASE	0x01BE
#define FN_F5_PRESS	0x013F
#define FN_F5_RELEASE	0x01BF
#define FN_F6_PRESS	0x0140
#define FN_F6_RELEASE	0x01C0
#define FN_F7_PRESS	0x0141
#define FN_F7_RELEASE	0x01C1
#define FN_F8_PRESS	0x0142
#define FN_F8_RELEASE	0x01C2
#define FN_F9_PRESS	0x0143
#define FN_F9_RELEASE	0x01C3

#define FN_BS_PRESS	0x010E
#define FN_BS_RELEASE	0x018E
#define FN_ESC_PRESS	0x0101
#define FN_ESC_RELEASE	0x0181
#define FN_KNJ_PRESS	0x0129
#define FN_KNJ_RELEASE	0x01A9

void
system_event(device_t dev, ACPI_HANDLE h)
{
    UINT32 key;

    while (1) {
        if (ACPI_FAILURE(get_hci_fn_key(h, &key))) {
            return;
        }

        switch(key) {
        case FN_ESC_RELEASE:
        case FN_F_RELEASE:
            fan_next(h);
            break; 
        case FN_F2_RELEASE:
            AcpiOsQueueForExecution(OSD_PRIORITY_LO,
                        acpi_toshiba_button_pressed_for_power_profile, dev);
            break;
        case FN_F3_RELEASE:
             AcpiOsQueueForExecution(OSD_PRIORITY_LO,
                        acpi_toshiba_button_notify_pressed_for_sleep, dev);
            break; 
        case FN_F5_RELEASE:
            video_output_next(h);
            break; 
        case FN_F6_RELEASE:
            lcd_brightness_down(h);
            break; 
        case FN_F7_RELEASE:
            lcd_brightness_up(h);
            break; 
        case FN_F8_RELEASE:
            lcd_backlight_next(h);
            break; 
#if 0
        default:
            printf("key=0x%x\n", key);
#endif
        }
    }
}

static void
acpi_toshiba_notify_handler(ACPI_HANDLE h, UINT32 notify, void *context)
{
    device_t dev;
    struct acpi_toshiba_softc	*sc;

    dev = (device_t)context;
    if ((sc = device_get_softc(dev)) == NULL)
        return;

    switch (notify) {
    case 0x80:
        system_event(dev, h);
        break;
    default:
        device_printf(dev,
                    "Toshiba Ex driver get Unknown Notify (0x%x)\n", notify);
        break;
    }
}

static void
acpi_toshiba_button_notify_pressed_for_sleep(void *arg)
{
    device_t dev;
    struct acpi_softc           *acpi_sc;

    dev = (device_t)arg;
    acpi_sc = acpi_device_get_parent_softc(dev);
    if (acpi_sc == NULL) {
        printf("cp100 acpi_device_get_parent_softc() failed\n");
        return_VOID;
    }

    acpi_eventhandler_sleep_button_for_sleep((void *)acpi_sc);
}

static void
acpi_toshiba_button_pressed_for_power_profile(void *arg)
{
    int status;

    status = power_profile_get_state();
    power_profile_set_state(status ? POWER_PROFILE_PERFORMANCE :
                                POWER_PROFILE_ECONOMY);
}


static device_method_t acpi_toshiba_methods[] = {
	DEVMETHOD(device_probe,		acpi_toshiba_probe),
	DEVMETHOD(device_attach,	acpi_toshiba_attach),
	{0, 0}
};

static driver_t acpi_toshiba_driver = {
	"acpi_toshiba",
	acpi_toshiba_methods,
	sizeof(struct acpi_toshiba_softc),
};

static devclass_t acpi_toshiba_devclass;
DRIVER_MODULE(acpi_toshiba, acpi, acpi_toshiba_driver, acpi_toshiba_devclass, 0, 0);
MODULE_DEPEND(acpi_toshiba, acpi, 100, 100, 100);
